'''
Created on Jan 24, 2016

@author: nicolas
'''

import os.path
from contextlib import contextmanager
import shutil
import threading

from lemoncheesecake.exceptions import LemonCheesecakeInternalError
from lemoncheesecake.consts import ATTACHEMENT_DIR, \
    LOG_LEVEL_DEBUG, LOG_LEVEL_ERROR, LOG_LEVEL_INFO, LOG_LEVEL_WARN
from lemoncheesecake.reporting import *
from lemoncheesecake import events
from lemoncheesecake.exceptions import ProgrammingError

__all__ = "log_debug", "log_info", "log_warn", "log_warning", "log_error", "log_url", "log_check", \
    "set_step", "end_step", "detached_step", \
    "prepare_attachment", "save_attachment_file", "save_attachment_content", \
    "add_report_info", "get_fixture"


_runtime = None  # singleton


def initialize_runtime(report_dir, report, fixture_registry):
    global _runtime
    _runtime = _Runtime(report_dir, report, fixture_registry)
    events.add_listener(_runtime)


def get_runtime():
    if not _runtime:
        raise LemonCheesecakeInternalError("Runtime is not initialized")
    return _runtime


class _Runtime:
    def __init__(self, report_dir, report, fixture_registry):
        self.report_dir = report_dir
        self.report = report
        self.fixture_registry = fixture_registry
        self.attachments_dir = os.path.join(self.report_dir, ATTACHEMENT_DIR)
        self.attachment_count = 0
        self.step_lock = False
        self.location = None
        self._local = threading.local()

    def set_step(self, description, force_lock=False, detached=False):
        if self.step_lock and not force_lock:
            return

        self._local.step = description
        events.fire(events.StepEvent(self.location, description, detached=detached))

    def end_step(self, step):
        events.fire(events.StepEndEvent(self.location, step))

    def _get_step(self, step):
        return step if step is not None else self._local.step

    def log(self, level, content, step=None):
        events.fire(events.LogEvent(self.location, self._get_step(step), level, content))

    def log_check(self, description, outcome, details, step=None):
        events.fire(events.CheckEvent(self.location, self._get_step(step), description, outcome, details))

    def log_url(self, url, description, step=None):
        events.fire(events.LogUrlEvent(self.location, self._get_step(step), url, description))

    @contextmanager
    def prepare_attachment(self, filename, description, step=None):
        attachment_filename = "%04d_%s" % (self.attachment_count + 1, filename)
        self.attachment_count += 1
        if not os.path.exists(self.attachments_dir):
            os.mkdir(self.attachments_dir)

        yield os.path.join(self.attachments_dir, attachment_filename)

        events.fire(events.LogAttachmentEvent(
            self.location, self._get_step(step), "%s/%s" % (ATTACHEMENT_DIR, attachment_filename), filename, description
        ))

    def save_attachment_file(self, filename, description, step=None):
        with self.prepare_attachment(os.path.basename(filename), description, step=step) as report_attachment_path:
            shutil.copy(filename, report_attachment_path)

    def save_attachment_content(self, content, filename, description, binary_mode, step=None):
        with self.prepare_attachment(filename, description, step=step) as report_attachment_path:
            with open(report_attachment_path, "wb") as fh:
                fh.write(content if binary_mode else content.encode("utf-8"))

    def get_fixture(self, name):
        try:
            scope = self.fixture_registry.get_fixture_scope(name)
        except KeyError:
            raise ProgrammingError("Fixture '%s' does not exist" % name)

        if scope != "session_prerun":
            raise ProgrammingError("Fixture '%s' has scope '%s' while only fixtures with scope 'session_prerun' can be retrieved" % (
                name, scope
            ))

        try:
            return self.fixture_registry.get_fixture_result(name)
        except AssertionError as excp:
            raise ProgrammingError(str(excp))


def set_step(description, force_lock=False, detached=False):
    """
    Set a new step.
    """
    get_runtime().set_step(description, force_lock=force_lock, detached=detached)


def end_step(step):
    """
    End a detached step.
    """
    get_runtime().end_step(step)


@contextmanager
def detached_step(description):
    set_step(description, detached=True)
    yield
    end_step(description)


def log_debug(content, step=None):
    """
    Log a debug level message.
    """
    get_runtime().log(LOG_LEVEL_DEBUG, content, step=step)


def log_info(content, step=None):
    """
    Log a info level message.
    """
    get_runtime().log(LOG_LEVEL_INFO, content, step=step)


def log_warning(content, step=None):
    """
    Log a warning level message.
    """
    get_runtime().log(LOG_LEVEL_WARN, content, step=step)


log_warn = log_warning


def log_error(content, step=None):
    """
    Log an error level message.
    """
    get_runtime().log(LOG_LEVEL_ERROR, content, step=step)


def log_check(description, outcome, details=None, step=None):
    get_runtime().log_check(description, outcome, details, step=step)


def prepare_attachment(filename, description=None, step=None):
    """
    Prepare a attachment using a pseudo filename and an optional description.
    The function returns the real filename on disk that will be used by the caller
    to write the attachment content.
    """
    return get_runtime().prepare_attachment(filename, description or filename, step=step)


def save_attachment_file(filename, description=None, step=None):
    """
    Save an attachment using an existing file (identified by filename) and an optional
    description. The given file will be copied.
    """
    get_runtime().save_attachment_file(filename, description or filename, step=step)


def save_attachment_content(content, filename, description=None, binary_mode=False, step=None):
    """
    Save a given content as attachment using pseudo filename and optional description.
    """
    get_runtime().save_attachment_content(content, filename, description or filename, binary_mode, step=step)


def log_url(url, description=None, step=None):
    """
    Log an URL.
    """
    get_runtime().log_url(url, description or url, step=step)


def get_fixture(name):
    """
    Return the corresponding fixture value. Only fixtures whose scope is session_prerun can be retrieved.
    """
    return get_runtime().get_fixture(name)


def add_report_info(name, value):
    report = get_runtime().report
    report.add_info(name, value)


def set_runtime_location(location):
    runtime = get_runtime()
    runtime.location = location



'''
Created on Sep 10, 2016

@author: nicolas
'''

import os
import os.path as osp
import sys
import glob
import fnmatch
import re
import imp
import warnings

from lemoncheesecake.exceptions import serialize_current_exception, ModuleImportError


def strip_py_ext(filename):
    return re.sub("\.py$", "", filename)


def get_py_files_from_dir(dir):
    return list(filter(
        lambda f: not os.path.basename(f).startswith("__"), glob.glob(os.path.join(dir, "*.py"))
    ))


def get_matching_files(patterns, excluding=[]):
    if type(patterns) not in (list, tuple):
        patterns = [patterns]
    if type(excluding) not in (list, tuple):
        excluding = [excluding]
    files = []
    for pattern in patterns:
        files.extend(glob.glob(pattern))
    if excluding:
        tmp = files[:] # iterate on copy to be able to alter files
        for file in tmp:
            for excluded in excluding:
                if fnmatch.fnmatch(file, excluded):
                    files.remove(file)
                    break
    return files


def import_module(mod_filename):
    mod_dir = osp.dirname(mod_filename)
    mod_name = strip_py_ext(osp.basename(mod_filename))

    ###
    # Find module
    ###
    sys.path.insert(0, mod_dir)
    try:
        with warnings.catch_warnings():
            # would raise a warning since module's directory does not need to have a __init__.py
            warnings.simplefilter("ignore", ImportWarning)
            fh, path, description = imp.find_module(mod_name)
    except ImportError:
        raise ModuleImportError(
            "Cannot find module '%s': %s" % (mod_filename, serialize_current_exception(show_stacktrace=True))
        )
    finally:
        del sys.path[0]

    ###
    # Load module
    ###
    package_name = "".join(osp.splitdrive(mod_dir)[1].split(osp.sep)[1:])
    try:
        mod = imp.load_module(package_name + mod_name, fh, path, description)
    except ImportError:
        raise ModuleImportError(
            "Cannot import file '%s': %s" % (mod_filename, serialize_current_exception(show_stacktrace=True))
        )
    except Exception:
        raise ModuleImportError(
            "Error while importing file '%s': %s" % (mod_filename, serialize_current_exception(show_stacktrace=True))
        )
    finally:
        fh.close()

    return mod



from __future__ import print_function

from lemoncheesecake.cli.utils import LEMONCHEESECAKE_VERSION
from lemoncheesecake.cli.command import Command


class VersionCommand(Command):
    def get_name(self):
        return "version"

    def get_description(self):
        return "Show lemoncheesecake version"

    def run_cmd(self, cli_args):
        print(LEMONCHEESECAKE_VERSION)



'''
Created on Jun 16, 2017

@author: nicolas
'''

import copy

from lemoncheesecake.utils import get_distincts_in_list
from lemoncheesecake.exceptions import CannotFindTreeNode


class TreeLocation(object):
    TEST_SESSION_SETUP = 0
    TEST_SESSION_TEARDOWN = 1
    SUITE_SETUP = 2
    SUITE_TEARDOWN = 3
    TEST = 4

    def __init__(self, node_type, node_hierarchy=None):
        self.node_type = node_type
        self.node_hierarchy = node_hierarchy
    
    @classmethod
    def in_test_session_setup(cls):
        return cls(cls.TEST_SESSION_SETUP)
    
    @classmethod
    def in_test_session_teardown(cls):
        return cls(cls.TEST_SESSION_TEARDOWN)
    
    @classmethod
    def in_suite_setup(cls, suite):
        return cls(cls.SUITE_SETUP, suite)
    
    @classmethod
    def in_suite_teardown(cls, suite):
        return cls(cls.SUITE_TEARDOWN, suite)

    @classmethod
    def in_test(cls, test):
        return cls(cls.TEST, test)


class BaseTreeNode(object):
    def __init__(self, name, description):
        self.parent_suite = None
        self.name = name
        self.description = description
        self.tags = []
        self.properties = {}
        self.links = []

    @property
    def hierarchy(self):
        if self.parent_suite is not None:
            for node in self.parent_suite.hierarchy:
                yield node
        yield self

    @property
    def hierarchy_depth(self):
        return len(list(self.hierarchy)) - 1

    @property
    def path(self):
        return ".".join([s.name for s in self.hierarchy])

    @property
    def hierarchy_paths(self):
        return map(lambda node: node.path, self.hierarchy)

    @property
    def hierarchy_descriptions(self):
        return map(lambda node: node.description, self.hierarchy)

    @property
    def hierarchy_tags(self):
        tags = []
        for node in self.hierarchy:
            tags.extend(node.tags)
        return get_distincts_in_list(tags)

    @property
    def hierarchy_properties(self):
        properties = {}
        for node in self.hierarchy:
            properties.update(node.properties)
        return properties

    @property
    def hierarchy_links(self):
        links = []
        for node in self.hierarchy:
            links.extend(node.links)
        return get_distincts_in_list(links)

    def pull_node(self):
        node = copy.copy(self)
        node.parent_suite = None
        return node

    def __str__(self):
        return self.path


class BaseTest(BaseTreeNode):
    pass


class BaseSuite(BaseTreeNode):
    def __init__(self, name, description):
        BaseTreeNode.__init__(self, name, description)
        self._tests = []
        self._suites = []

    def add_test(self, test):
        test.parent_suite = self
        self._tests.append(test)

    def get_tests(self):
        return self._tests

    def add_suite(self, suite):
        suite.parent_suite = self
        self._suites.append(suite)

    def get_suites(self, include_empty_suites=False):
        if include_empty_suites:
            return self._suites
        else:
            return list(filter(lambda suite: not suite.is_empty(), self._suites))

    def is_empty(self):
        if len(self.get_tests()) != 0:
            return False

        for sub_suite in self.get_suites():
            if not sub_suite.is_empty():
                return False

        return True

    def pull_node(self):
        node = BaseTreeNode.pull_node(self)
        node._tests = []
        node._suites = []
        return node


def flatten_suites(suites):
    for suite in suites:
        yield suite
        for sub_suite in flatten_suites(suite.get_suites()):
            yield sub_suite


def flatten_tests(suites):
    for suite in flatten_suites(suites):
        for test in suite.get_tests():
            yield test


def _normalize_node_hierarchy(value):
    if isinstance(value, BaseTreeNode):
        return [p.name for p in value.hierarchy]
    else:
        return value.split(".")


def get_suite_by_name(suites, suite_name):
    try:
        return next(s for s in suites if s.name == suite_name)
    except StopIteration:
        raise CannotFindTreeNode("Cannot find suite named '%s'" % suite_name)


def _find_suite(suites, hierarchy):
    lookup_suites = suites
    lookup_suite = None
    for lookup_suite_name in hierarchy:
        lookup_suite = get_suite_by_name(lookup_suites, lookup_suite_name)
        lookup_suites = lookup_suite.get_suites(include_empty_suites=True)
    if lookup_suite is None:
        raise CannotFindTreeNode("Cannot find suite named %s" % hierarchy)

    return lookup_suite


def find_suite(suites, hierarchy):
    return _find_suite(suites, _normalize_node_hierarchy(hierarchy))


def get_test_by_name(suite, test_name):
    try:
        return next(t for t in suite.get_tests() if t.name == test_name)
    except StopIteration:
        raise CannotFindTreeNode("Cannot find test named '%s'" % test_name)


def _find_test(suites, hierarchy):
    lookup_suite = _find_suite(suites, hierarchy[:-1])
    return get_test_by_name(lookup_suite, hierarchy[-1])


def find_test(suites, hierarchy):
    return _find_test(suites, _normalize_node_hierarchy(hierarchy))



'''
Created on Sep 8, 2016

@author: nicolas
'''

from lemoncheesecake.exceptions import InvalidMetadataError, ProgrammingError

__all__ = ("MetadataPolicy",)

class MetadataPolicy:
    def __init__(self):
        self._properties = {}
        self._tags = {}
        self._disallow_unknown_properties = False
        self._disallow_unknown_tags = False

    def has_constraints(self):
        return len(self._properties) != 0 or len(self._tags) != 0 or self._disallow_unknown_properties or self._disallow_unknown_tags

    def _get_rule_application(self, on_test, on_suite):
        if on_test == None and on_suite == None:
            return True, False
        if not on_test and not on_suite:
            raise ProgrammingError("either on_test or on_suite need to be True")
        return bool(on_test), bool(on_suite)

    def add_property_rule(self, prop_name, accepted_values=None, on_test=None, on_suite=None, required=False):
        """
        Declare a property rule with:
        - the property name
        - an optional list of accepted values
        - is the property available for tests
        - is the property available for suites
        - if the property is required
        If neither on_test or on_suite argument are set, then the property is only available for tests.
        """
        on_test, on_suite = self._get_rule_application(on_test, on_suite)
        self._properties[prop_name] = {
            "values": accepted_values,
            "on_test": on_test,
            "on_suite": on_suite,
            "required": required
        }

    def disallow_unknown_properties(self):
        """
        Disallow unknown properties for tests and suites.
        """
        self._disallow_unknown_properties = True

    def add_tag_rule(self, tag_name, on_test=None, on_suite=None):
        """
        Declare a tag constraint with:
        - the tag name
        - is the tag available for tests
        - is the tag available for suites
        If neither on_test or on_suite argument are set, then the tag is only available for tests.
        """
        tag_names = tag_name if type(tag_name) in (list, tuple) else [tag_name]
        for tag_name in tag_names:
            on_test, on_suite = self._get_rule_application(on_test, on_suite)
            self._tags[tag_name] = {
                "on_test": on_test,
                "on_suite": on_suite
            }

    def disallow_unknown_tags(self):
        """
        Disallow unknown tags for tests and suites.
        """
        self._disallow_unknown_tags = True

    def _check_compliance(self, obj, obj_type,
                          available_properties, forbidden_properties,
                          available_tags, forbidden_tags):
        # check unknown properties
        if self._disallow_unknown_properties:
            for property_name in obj.properties.keys():
                if not property_name in available_properties:
                    help_msg = "available are %s" % ", ".join(available_properties.keys()) \
                        if available_properties else "no property is available"
                    raise InvalidMetadataError(
                        "In %s '%s', the property '%s' is not supported (%s)" % (
                        obj_type, obj.path, property_name, help_msg
                    ))

        # check forbidden properties
        for property_name in obj.properties.keys():
            if property_name in forbidden_properties:
                raise InvalidMetadataError(
                    "In %s '%s', the property '%s' is not accepted on a %s" % (
                        obj_type, obj.path, property_name, obj_type
                    )
                )

        # check required properties
        for required_property in filter(lambda p: available_properties[p]["required"], available_properties.keys()):
            if not required_property in obj.properties.keys():
                raise InvalidMetadataError(
                    "In %s '%s', the mandatory property '%s' is missing" % (
                    obj_type, obj.path, required_property
                ))

        # check properties allowed values
        for name, value in obj.properties.items():
            if name not in available_properties:
                continue
            if available_properties[name]["values"] and not value in available_properties[name]["values"]:
                raise InvalidMetadataError(
                    "In %s '%s', value '%s' of property '%s' is not among accepted values: %s" % (
                    obj_type, obj.path, value, name, available_properties[name]["values"]
                ))

        # check unknown tags
        if self._disallow_unknown_tags:
            for tag in obj.tags:
                if tag not in available_tags.keys():
                    help_msg = "available are %s" % ", ".join(available_tags.keys()) \
                        if available_tags else "no property is available"
                    raise InvalidMetadataError(
                        "In %s '%s', the tag '%s' is not supported (%s)" % (
                        obj_type, obj.path, tag, help_msg
                    ))

        # check forbidden tags
        for tag in obj.tags:
            if tag in forbidden_tags:
                raise InvalidMetadataError(
                    "In %s '%s', the tag '%s' is not accepted on a %s" % (
                        obj_type, obj.path, tag, obj_type
                    )
                )

    def check_test_compliance(self, test):
        """
        Check if the test complies to the metadata policy.
        Raise InvalidMetadataError if not compliant.
        """
        self._check_compliance(
            test, "test",
            {prop_name: p for prop_name, p in self._properties.items() if p["on_test"]},
            [prop_name for prop_name, p in self._properties.items() if not p["on_test"]],
            {tag_name: t for tag_name, t in self._tags.items() if t["on_test"]},
            [tag_name for tag_name, t in self._tags.items() if not t["on_test"]]
        )

    def check_suite_compliance(self, suite, recursive=True):
        """
        Check if the suite complies to the metadata policy.
        If recursive if set to True (which is the default), then suite tests and sub suites are also checked.
        Raise InvalidMetadataError if not compliant.
        """
        self._check_compliance(
            suite, "suite",
            {prop_name: p for prop_name, p in self._properties.items() if p["on_suite"]},
            [prop_name for prop_name, p in self._properties.items() if not p["on_suite"]],
            {tag_name: t for tag_name, t in self._tags.items() if t["on_suite"]},
            [tag_name for tag_name, t in self._tags.items() if not t["on_suite"]]
        )
        if not recursive:
            return

        for test in suite.get_tests():
            self.check_test_compliance(test)

        for sub_suite in suite.get_suites():
            self.check_suite_compliance(sub_suite, recursive=True)

    def check_suites_compliance(self, suites):
        """
        Check if the suites comply to the metadata policy.
        Raise InvalidMetadataError if not compliant.
        """
        for suite in suites:
            self.check_suite_compliance(suite, recursive=True)



from .value import *
from .string import *
from .list_ import *
from .dict_ import *
from .types_ import *
from .composites import *

__all__ = list(filter(lambda sym: not sym.startswith("_"), dir()))


'''
Created on Sep 8, 2016

@author: nicolas
'''

import inspect
import copy

from lemoncheesecake.suite.core import InjectedFixture
from lemoncheesecake.exceptions import ProgrammingError

__all__ = "add_test_into_suite", "add_test_in_suite", "add_tests_in_suite", "get_metadata", \
    "suite", "test", "tags", "prop", "link", "disabled", "conditional", "hidden", "inject_fixture"


class Metadata:
    _next_rank = 1

    def __init__(self):
        self.is_test = False
        self.is_suite = False
        self.name = None
        self.description = None
        self.properties = {}
        self.tags = []
        self.links = []
        self.rank = 0
        self.disabled = False
        self.condition = None


def get_metadata_next_rank():
    rank = Metadata._next_rank
    Metadata._next_rank += 1
    return rank


def add_test_into_suite(test, suite):
    if not hasattr(suite, "_lccgeneratedtests"):
        suite._lccgeneratedtests = []
    suite._lccgeneratedtests.append(test)


def add_test_in_suite(test, suite, before_test=None, after_test=None):
    """
    backward-compatibility function: use add_test_into_suite instead
    before_test and after_test arguments are simply ignored
    """
    add_test_into_suite(test, suite)


def add_tests_in_suite(tests, suite, before_test=None, after_test=None):
    """
    backward-compatibility function: loop over tests and use add_test_into_suite instead
    before_test and after_test arguments are simply ignored
    """
    for test in tests:
        add_test_into_suite(test, suite)


_objects_with_metadata = []
def get_metadata(obj):
    global _objects_with_metadata

    if hasattr(obj, "_lccmetadata"):
        if obj not in _objects_with_metadata: # metadata comes from the superclass
            obj._lccmetadata = copy.deepcopy(obj._lccmetadata)
        return obj._lccmetadata
    else:
        obj._lccmetadata = Metadata()
        _objects_with_metadata.append(obj)
        return obj._lccmetadata


def suite(description, name=None, rank=None):
    """Decorator, mark a class as a suite class"""
    def wrapper(klass):
        if not inspect.isclass(klass):
            raise ProgrammingError("%s is not a class (suite decorator can only be used on a class)" % klass)
        md = get_metadata(klass)
        md.is_suite = True
        md.rank = rank if rank != None else get_metadata_next_rank()
        md.name = name or klass.__name__
        md.description = description
        return klass
    return wrapper


def test(description, name=None):
    """Decorator, make a method as a test method"""
    def wrapper(func):
        if not inspect.isfunction(func):
            raise ProgrammingError("%s is not a function (test decorator can only be used on a function)" % func)
        md = get_metadata(func)
        md.is_test = True
        md.rank = get_metadata_next_rank()
        md.name = name or func.__name__
        md.description = description
        return func
    return wrapper


def tags(*tag_names):
    """Decorator, add tags to a test or a suite"""
    def wrapper(obj):
        md = get_metadata(obj)
        md.tags.extend(tag_names)
        return obj
    return wrapper


def prop(key, value):
    """Decorator, add a property (key/value) to a test or a suite"""
    def wrapper(obj):
        md = get_metadata(obj)
        md.properties[key] = value
        return obj
    return wrapper


def link(url, name=None):
    """Decorator, set a link (with an optional friendly name) to a test or a suite"""
    def wrapper(obj):
        md = get_metadata(obj)
        md.links.append((url, name))
        return obj
    return wrapper


def disabled():
    """Decorator, disable a test or a suite"""
    def wrapper(obj):
        md = get_metadata(obj)
        md.disabled = True
        return obj
    return wrapper


def conditional(condition):
    """Decorator, the test or suite will only appear if the given condition callback return a true value"""
    def wrapper(obj):
        md = get_metadata(obj)
        md.condition = condition
        return obj
    return wrapper


def hidden():
    return conditional(lambda _: False)


def inject_fixture(fixture_name=None):
    return InjectedFixture(fixture_name)


import os.path

from lemoncheesecake.project import SimpleProjectConfiguration


project_dir = os.path.dirname(__file__)
project = SimpleProjectConfiguration(
    suites_dir=os.path.join(project_dir, "suites"),
    fixtures_dir=os.path.join(project_dir, "fixtures"),
)


from __future__ import print_function

import os
import sys
import traceback
import mimetypes

try:
    from reportportal_client import ReportPortalServiceAsync
    REPORT_PORTAL_CLIENT_IS_AVAILABLE = True
except ImportError:
    REPORT_PORTAL_CLIENT_IS_AVAILABLE = False

from lemoncheesecake.reporting.backend import ReportingBackend, ReportingSession
from lemoncheesecake.exceptions import UserError


def make_time(t):
    return str(int(t * 1000))


def convert_properties_into_tags(props):
    return ["%s_%s" % (name, value) for name, value in props.items()]


def convert_links_into_tags(links):
    return [link[1] or link[0] for link in links]


class ReportPortalReportingSession(ReportingSession):
    def __init__(self, url, auth_token, project, launch_name, launch_description, report_dir, report):
        self.service = ReportPortalServiceAsync(
            endpoint=url, project=project, token=auth_token, error_handler=self._handle_rp_error
        )
        self.launch_name = launch_name
        self.launch_description = launch_description
        self.report_dir = report_dir
        self.report = report
        self._rp_exc_info = None
        self._current_suite_test_statuses = []

    def _handle_rp_error(self, exc_info):
        self._rp_exc_info = exc_info
        return False  # stop on error

    def _has_rp_error(self):
        return self._rp_exc_info is not None

    def _show_rp_error(self):
        print(
            "Got the following exception using ReportPortal, "
            "test results have not been properly synced:",
            file=sys.stderr
        )
        traceback.print_exception(*self._rp_exc_info, file=sys.stderr)

    def _end_current_test_item(self, end_time, status):
        self.service.finish_test_item(end_time=make_time(end_time), status=status)

    def _start_test_item(self, item_type, start_time, name, description, wrapped=False):
        if wrapped:
            self.service.start_test_item(
                item_type="SUITE", start_time=make_time(start_time),
                name=name, description=description
            )
        self.service.start_test_item(
            item_type=item_type, start_time=make_time(start_time),
            name=name, description=description
        )

    def _end_test_item(self, end_time, outcome, wrapped=False):
        status = "passed" if outcome else "failed"
        if wrapped:
            self._end_current_test_item(end_time, status=status)
        self._end_current_test_item(end_time, status=status)

    def on_test_session_start(self, event):
        if self._has_rp_error():
            return

        self.service.start_launch(
            name=self.launch_name, description=self.launch_description, start_time=make_time(event.time)
        )

    def on_test_session_end(self, event):
        if self._has_rp_error():
            self._show_rp_error()
        else:
            self.service.finish_launch(end_time=make_time(event.time))
            self.service.terminate()

            if self._has_rp_error():
                self._show_rp_error()

    def on_test_session_setup_start(self, event):
        if self._has_rp_error():
            return

        self._start_test_item(
            item_type="BEFORE_CLASS", start_time=event.time,
            name="session_setup", description="Test Session Setup",
            wrapped=True
        )

    def on_test_session_setup_end(self, event):
        if self._has_rp_error():
            return

        self._end_test_item(
            event.time,
            False if self.report.test_session_setup and self.report.test_session_setup.has_failure() else True,
            wrapped=True
        )

    def on_test_session_teardown_start(self, event):
        if self._has_rp_error():
            return

        self._start_test_item(
            item_type="AFTER_CLASS", start_time=event.time,
            name="session_teardown", description="Test Session Teardown",
            wrapped=True
        )

    def on_test_session_teardown_end(self, event):
        if self._has_rp_error():
            return

        self._end_test_item(event.time, event.teardown_outcome, wrapped=True)

    def on_suite_start(self, event):
        if self._has_rp_error():
            return

        suite = event.suite
        self.service.start_test_item(
            item_type="SUITE", start_time=make_time(event.time),
            name=suite.name, description=suite.description,
            tags=suite.tags +
                convert_properties_into_tags(suite.properties) +
                convert_links_into_tags(suite.links),
        )

    def on_suite_end(self, event):
        if self._has_rp_error():
            return

        self._end_current_test_item(event.time, status="passed")

    def on_suite_setup_start(self, event):
        if self._has_rp_error():
            return

        self._start_test_item(
            item_type="BEFORE_CLASS", start_time=event.time,
            name="suite_setup", description="Suite Setup",
            wrapped=len(event.suite.get_suites()) > 0
        )

    def on_suite_setup_end(self, event):
        if self._has_rp_error():
            return

        self._end_test_item(event.time, outcome=event.time, wrapped=len(event.suite.get_suites()) > 0)

    def on_suite_teardown_start(self, event):
        if self._has_rp_error():
            return

        self._start_test_item(
            item_type="AFTER_CLASS", start_time=event.time,
            name="suite_teardown", description="Suite Teardown",
            wrapped=len(event.suite.get_suites()) > 0
        )

    def on_suite_teardown_end(self, event):
        if self._has_rp_error():
            return

        self._end_test_item(event.time, outcome=event.teardown_outcome, wrapped=len(event.suite.get_suites()) > 0)

    def on_test_start(self, event):
        if self._has_rp_error():
            return

        test = event.test
        self.service.start_test_item(
            item_type="TEST", start_time=make_time(event.time),
            name=test.name, description=test.description,
            tags=test.tags +
                 convert_properties_into_tags(test.properties) +
                 convert_links_into_tags(test.links)
        )

    def on_test_end(self, event):
        if self._has_rp_error():
            return

        test_data = self.report.get_test(event.test)
        self._end_current_test_item(event.time, test_data.status)

    def _bypass_test(self, test, status, time):
        if self._has_rp_error():
            return

        self.service.start_test_item(
            item_type="TEST", start_time=make_time(time),
            name=test.name, description=test.description, tags=test.tags,
        )
        self._end_current_test_item(time, status=status)

    def on_test_skipped(self, event):
        if self._has_rp_error():
            return

        self._bypass_test(event.test, "skipped", event.time)

    def on_disabled_test(self, event):
        # do not log disabled test, moreover it seems that the is not corresponding status in ReportPortal
        pass

    def on_step(self, event):
        if self._has_rp_error():
            return

        self.service.log(make_time(event.time), "--- STEP: %s ---" % event.step_description, "INFO")

    def on_log(self, event):
        if self._has_rp_error():
            return

        self.service.log(make_time(event.time), event.log_message, event.log_level)

    def on_check(self, event):
        if self._has_rp_error():
            return

        message = "%s => %s" % (event.check_description, "OK" if event.check_outcome else "NOT OK")
        if event.check_details is not None:
            message += "\nDetails: %s" % event.check_details
        self.service.log(make_time(event.time), message, "INFO" if event.check_outcome else "ERROR")

    def on_log_attachment(self, event):
        if self._has_rp_error():
            return

        abspath = os.path.join(self.report_dir, event.attachment_path)
        with open(abspath, "rb") as fh:
            self.service.log(make_time(event.time), event.attachment_description, "INFO", attachment={
                "name": event.attachment_filename,
                "data": fh.read(),
                "mime": mimetypes.guess_type(abspath)[0] or "application/octet-stream"
            })

    def on_log_url(self, event):
        if self._has_rp_error():
            return

        if event.url_description and event.url_description != event.url:
            message = "%s: %s" % (event.url_description, event.url)
        else:
            message = event.url
        self.service.log(make_time(event.time), message, "INFO")


class ReportPortalBackend(ReportingBackend):
    name = "reportportal"

    def is_available(self):
        return REPORT_PORTAL_CLIENT_IS_AVAILABLE

    def create_reporting_session(self, report_dir, report):
        try:
            url = os.environ["RP_URL"]
            auth_token = os.environ["RP_AUTH_TOKEN"]
            project = os.environ["RP_PROJECT"]
            launch_name = os.environ.get("RP_LAUNCH_NAME", "Test Run")
            launch_description = os.environ.get("RP_LAUNCH_DESCRIPTION", None)
        except KeyError as excp:
            raise UserError("ReportPortal reporting backend, cannot get environment variable %s" % excp)

        return ReportPortalReportingSession(
            url, auth_token, project, launch_name, launch_description, report_dir, report
        )



'''
Created on Jan 7, 2017

@author: nicolas
'''

import inspect

from lemoncheesecake.importer import import_module, get_matching_files, get_py_files_from_dir
from lemoncheesecake.exceptions import FixtureError, ProgrammingError
from lemoncheesecake.utils import get_distincts_in_list, get_callable_args

__all__ = (
    "fixture",
    "load_fixtures_from_func", "load_fixtures_from_file",
    "load_fixtures_from_files", "load_fixtures_from_directory",
    "inject_fixture"
)

FORBIDDEN_FIXTURE_NAMES = ("fixture_name", )
SCOPE_LEVELS = {
    "test": 1,
    "suite": 2,
    "session": 3,
    "session_prerun": 4
}


class FixtureInfo:
    def __init__(self, names, scope):
        self.names = names
        self.scope = scope


def fixture(names=None, scope="test"):
    def wrapper(func):
        if scope not in SCOPE_LEVELS.keys():
            raise ProgrammingError("Invalid fixture scope '%s' in fixture function '%s'" % (scope, func.__name__))

        setattr(func, "_lccfixtureinfo", FixtureInfo(names, scope))
        return func

    return wrapper


class BaseFixture:
    def is_builtin(self):
        return False

    def get_scope_level(self):
        return {
            "test": 1,
            "suite": 2,
            "session": 3,
            "session_prerun": 4
        }[self.scope]

    def is_executed(self):
        return hasattr(self, "_result")

    def teardown(self):
        pass

    def reset(self):
        pass


class Fixture(BaseFixture):
    def __init__(self, name, func, scope, params):
        self.name = name
        self.func = func
        self.scope = scope
        self.params = params
        self._generator = None

    def execute(self, params={}):
        assert not self.is_executed(), "fixture '%s' has already been executed" % self.name
        for param_name in params.keys():
            assert param_name in self.params

        result = self.func(**params)
        if inspect.isgenerator(result):
            self._generator = result
            self._result = next(result)
        else:
            self._result = result

    def get_result(self):
        assert self.is_executed(), "fixture '%s' has not been executed" % self.name
        return self._result

    def teardown(self):
        assert self.is_executed(), "fixture '%s' has not been executed" % self.name
        delattr(self, "_result")
        if self._generator:
            try:
                next(self._generator)
            except StopIteration:
                self._generator = None
            else:
                raise FixtureError("The fixture yields more than once, only one yield is supported")


class BuiltinFixture(BaseFixture):
    def __init__(self, name, value):
        self.name = name
        self.scope = "session_prerun"
        self.params = []
        self._value = value

    def is_builtin(self):
        return True

    def execute(self, params={}):
        self._result = self._value() if callable(self._value) else self._value

    def get_result(self):
        return self._result


class FixtureRegistry:
    def __init__(self):
        self._fixtures = {}

    def add_fixture(self, fixture):
        if fixture.name in self._fixtures and self._fixtures[fixture.name].is_builtin():
            raise FixtureError("'%s' is a builtin fixture name" % fixture.name)
        self._fixtures[fixture.name] = fixture

    def add_fixtures(self, fixtures):
        for fixture in fixtures:
            self.add_fixture(fixture)

    def get_fixture(self, name):
        return self._fixtures[name]

    def _get_fixture_dependencies(self, name, orig_fixture):
        fixture_params = [p for p in self._fixtures[name].params if p != "fixture_name"]
        if orig_fixture and orig_fixture in fixture_params:
            raise FixtureError("Fixture '%s' has a circular dependency on fixture '%s'" % (orig_fixture, name))

        dependencies = []
        for param in fixture_params:
            if param not in self._fixtures:
                raise FixtureError("Fixture '%s' used by fixture '%s' does not exist" % (param, name))
            dependencies.extend(self._get_fixture_dependencies(param, orig_fixture if orig_fixture else name))
        dependencies.extend(fixture_params)

        return dependencies

    def get_fixture_dependencies(self, name):
        dependencies = self._get_fixture_dependencies(name, None)
        return get_distincts_in_list(dependencies)

    def filter_fixtures(self, base_names=[], scope=None, is_executed=None, with_dependencies=False):
        def do_filter_fixture(fixture):
            if scope != None and fixture.scope != scope:
                return False
            if is_executed != None and fixture.is_executed() != is_executed:
                return False
            return True

        names = base_names if base_names else self._fixtures.keys()
        fixtures = filter(do_filter_fixture, [self._fixtures[name] for name in names])
        return [f.name for f in fixtures]

    def check_dependencies(self):
        """
        Checks for:
        - missing dependencies
        - circular dependencies
        - scope incoherence
        - forbidden fixture name
        raises FixtureError if a check fails
        """
        # first, check for forbidden fixture name
        for fixture_name in self._fixtures.keys():
            if fixture_name in FORBIDDEN_FIXTURE_NAMES:
                raise FixtureError("Fixture name '%s' is forbidden" % fixture_name)

        # second, check for missing & circular dependencies
        for fixture_name in self._fixtures.keys():
            self.get_fixture_dependencies(fixture_name)

        # third, check fixture scope compliance with their direct fixture dependencies
        for fixture in self._fixtures.values():
            dependency_fixtures = [self._fixtures[param] for param in fixture.params if param != "fixture_name"]
            for dependency_fixture in dependency_fixtures:
                if dependency_fixture.get_scope_level() < fixture.get_scope_level():
                    raise FixtureError("Fixture '%s' with scope '%s' is incompatible with scope '%s' of fixture '%s'" % (
                        fixture.name, fixture.scope, dependency_fixture.scope, dependency_fixture.name
                    ))

    def check_fixtures_in_test(self, test):
        for fixture in test.get_fixtures():
            if fixture not in self._fixtures:
                raise FixtureError("Unknown fixture '%s' used in test '%s'" % (fixture, test.path))

    def check_fixtures_in_suite(self, suite):
        for fixture in suite.get_fixtures():
            if fixture not in self._fixtures:
                raise FixtureError("Suite '%s' uses an unknown fixture '%s'" % (suite.path, fixture))
            if self._fixtures[fixture].get_scope_level() < SCOPE_LEVELS["suite"]:
                raise FixtureError("Suite '%s' uses fixture '%s' which has an incompatible scope" % (
                    suite.path, fixture
                ))

        for test in suite.get_tests():
            self.check_fixtures_in_test(test)

        for sub_suite in suite.get_suites():
            self.check_fixtures_in_suite(sub_suite)

    def check_fixtures_in_suites(self, suites):
        for suite in suites:
            self.check_fixtures_in_suite(suite)

    def get_fixture_scope(self, name):
        return self._fixtures[name].scope

    def execute_fixture(self, name):
        fixture = self._fixtures[name]
        params = {}
        for param in fixture.params:
            if param == "fixture_name":
                params["fixture_name"] = name
            else:
                dependency_fixture = self._fixtures[param]
                if not dependency_fixture.is_executed():
                    self.execute_fixture(dependency_fixture.name)
                params[dependency_fixture.name] = dependency_fixture.get_result()
        fixture.execute(params)

    def get_fixture_result(self, name):
        return self._fixtures[name].get_result()

    def is_fixture_executed(self, name):
        return self._fixtures[name].is_executed()

    def get_fixture_results(self, names):
        results = {}
        for name in names:
            results[name] = self.get_fixture_result(name)
        return results

    def teardown_fixture(self, name):
        self._fixtures[name].teardown()


def load_fixtures_from_func(func):
    assert hasattr(func, "_lccfixtureinfo")
    names = func._lccfixtureinfo.names
    if not names:
        names = [func.__name__]
    scope = func._lccfixtureinfo.scope
    args = get_callable_args(func)
    return [Fixture(name, func, scope, args) for name in names]


def load_fixtures_from_file(filename):
    mod = import_module(filename)

    fixtures = []
    for sym_name in dir(mod):
        sym = getattr(mod, sym_name)
        if hasattr(sym, "_lccfixtureinfo"):
            fixtures.extend(load_fixtures_from_func(sym))

    return fixtures


def load_fixtures_from_files(patterns, excluding=[]):
    """
    Import fixtures from a list of files:
    - patterns: a mandatory list (a simple string can also be used instead of a single element list)
      of files to import; the wildcard '*' character can be used
    - exclude: an optional list (a simple string can also be used instead of a single element list)
      of elements to exclude from the expanded list of files to import
    Example: load_suites_from_files("test_*.py")
    """
    fixtures = []
    for file in get_matching_files(patterns, excluding):
        fixtures.extend(load_fixtures_from_file(file))
    return fixtures


def load_fixtures_from_directory(dir):
    fixtures = []
    for file in get_py_files_from_dir(dir):
        fixtures.extend(load_fixtures_from_file(file))
    return fixtures



'''
Created on Feb 14, 2017

@author: nicolas
'''

from __future__ import print_function

from lemoncheesecake.cli.display import print_table
from lemoncheesecake.cli.command import Command
from lemoncheesecake.testtree import flatten_tests
from lemoncheesecake.project import load_project


class FixturesCommand(Command):
    def get_name(self):
        return "fixtures"

    def get_description(self):
        return "Show the fixtures available in the project"

    def add_cli_args(self, cli_parser):
        group = cli_parser.add_argument_group("Display")
        self.add_color_cli_args(group)
        group.add_argument("--verbose", "-v", action="store_true", help="Show extra fixture information")

    def show_fixtures(self, scope, fixtures, used_by_tests, used_by_fixtures, verbose):
        lines = []
        ordered_fixtures = sorted(
            fixtures,
            key=lambda f: used_by_fixtures.get(f.name, 0) + used_by_tests.get(f.name, 0),
            reverse=True
        )
        for fixt in ordered_fixtures:
            lines.append([
                self.bold(fixt.name), ", ".join(fixt.params or "-"),
                used_by_fixtures.get(fixt.name, 0), used_by_tests.get(fixt.name, 0)
            ])
        print_table("Fixture with scope %s" % self.bold(scope), ["Fixture", "Dependencies", "Used by fixtures", "Used by tests"], lines)

    def run_cmd(self, cli_args):
        self.process_color_cli_args(cli_args)

        project = load_project()
        suites = project.get_suites()
        fixtures = project.get_fixtures()

        fixtures_by_scope = {}
        for fixt in fixtures:
            if fixt.scope in fixtures_by_scope:
                fixtures_by_scope[fixt.scope].append(fixt)
            else:
                fixtures_by_scope[fixt.scope] = [fixt]

        used_by_tests = {}
        for test in flatten_tests(suites):
            for fixt_name in test.get_fixtures():
                used_by_tests[fixt_name] = used_by_tests.get(fixt_name, 0) + 1

        used_by_fixtures = {}
        for fixt in fixtures:
            for param in fixt.params:
                used_by_fixtures[param] = used_by_fixtures.get(param, 0) + 1

        for scope in "session_prerun", "session", "suite", "test":
            self.show_fixtures(scope, fixtures_by_scope.get(scope, []), used_by_tests, used_by_fixtures, cli_args.verbose)
            print()

        return 0


'''
Created on May 2, 2017

@author: nicolas
'''

from lemoncheesecake.matching.base import MatchExpected, match_failure, match_success, match_result, got_value, to_have, to_be, serialize_values
from lemoncheesecake.matching.matchers.composites import is_

__all__ = ("has_item", "has_values", "has_only_values", "is_in")


class HasItem(MatchExpected):
    def description(self, conjugate=False):
        return "%s an item whose value %s" % (to_have(conjugate), self.expected.description(conjugate=True))

    def short_description(self, conjugate=False):
        return "%s an item whose value %s" % (to_have(conjugate), self.expected.short_description(conjugate=True))

    def matches(self, actual):
        for item in actual:
            result = self.expected.matches(item)
            if result.is_success():
                return result
        return match_failure("no matching item")


def has_item(expected):
    "Test if iterable has item matching expected"
    return HasItem(is_(expected))


class HasValues(MatchExpected):
    def description(self, conjugate=False):
        return "%s values %s" % (to_have(conjugate), serialize_values(self.expected))

    def matches(self, actual):
        missing = []
        for expected in self.expected:
            if expected not in actual:
                missing.append(expected)

        if missing:
            return match_failure("Missing values: %s" % serialize_values(missing))
        else:
            return match_success(got_value(actual))


def has_values(values):
    "Test if iterable contains at least the given values"
    return HasValues(values)


class HasOnlyValues(MatchExpected):
    def description(self, conjugate=False):
        return "%s only values %s" % (to_have(conjugate), serialize_values(self.expected))

    def matches(self, actual):
        expected = list(self.expected)
        extra = []
        for value in actual:
            if value in expected:
                expected.remove(value)
            else:
                extra.append(value)

        if len(expected) == 0 and len(extra) == 0:
            return match_success(got_value(actual))
        else:
            details = []
            if len(expected) > 0:
                details.append("Missing values: %s" % serialize_values(expected))
            if len(extra) > 0:
                details.append("Extra values: %s" % serialize_values(extra))
            return match_failure("; ".join(details))


def has_only_values(values):
    "Test if iterable only contains the given values"
    return HasOnlyValues(values)


class IsIn(MatchExpected):
    def description(self, conjugate=False):
        return "%s in %s" % (to_be(conjugate), serialize_values(self.expected))

    def matches(self, actual):
        return match_result(actual in self.expected, got_value(actual))


def is_in(expected):
    return IsIn(expected)



from lemoncheesecake.cli.command import Command
from lemoncheesecake.cli.utils import auto_detect_reporting_backends, add_report_path_cli_arg, get_report_path
from lemoncheesecake.reporting import load_report
from lemoncheesecake.reporting.backends.console import display_report_suites
from lemoncheesecake.filter import add_report_filter_cli_args, make_report_filter, filter_suites


class ReportCommand(Command):
    def get_name(self):
        return "report"

    def get_description(self):
        return "Display a report"

    def add_cli_args(self, cli_parser):
        group = cli_parser.add_argument_group("Show report")
        add_report_path_cli_arg(group)
        add_report_filter_cli_args(cli_parser, no_positional_argument=True)

    def run_cmd(self, cli_args):
        report_path = get_report_path(cli_args)
        report = load_report(report_path, auto_detect_reporting_backends())
        suites = filter_suites(report.get_suites(), make_report_filter(cli_args))

        display_report_suites(suites)

        return 0



'''
Created on Mar 12, 2017

@author: nicolas
'''

import sys

import termcolor


class Command:
    def __init__(self):
        self.force_color = False

    def get_name(self):
        pass

    def get_description(self):
        return None

    def add_cli_args(self, cli_parser):
        pass

    def run_cmd(self, cli_args):
        pass

    def add_color_cli_args(self, cli_parser):
        cli_parser.add_argument("--color", action="store_true", help="Force color printing when command output is redirected")

    def process_color_cli_args(self, cli_args):
        self.force_color = cli_args.color

    def colored(self, text, color=None, on_color=None, attrs=None):
        if sys.stdout.isatty() or self.force_color:
            return termcolor.colored(text, color, on_color, attrs)
        else:
            return text

    def bold(self, text):
        return self.colored(text, attrs=["bold"])



'''
Created on Apr 1, 2017

@author: nicolas
'''

from lemoncheesecake.exceptions import method_not_implemented
from lemoncheesecake.matching.base import Matcher, match_failure, match_success, got_value, to_have, serialize_value
from lemoncheesecake.matching.matchers.composites import is_

__all__ = (
    "has_entry",
)


class EntryMatcher(object):
    def description(self):
        method_not_implemented("description", self)

    def get_entry(self, actual):
        """Return the value of dict corresponding to entry matching or raise KeyError if entry is not found"""
        method_not_implemented("get_entry", self)


class KeyPathMatcher(EntryMatcher):
    """Dict lookup through a list of key, each key represent a level of depth of the dict"""
    def __init__(self, path):
        self.path = path

    def description(self):
        return " -> ".join(map(serialize_value, self.path))

    def get_entry(self, actual):
        d = actual
        for key in self.path:
            try:
                d = d[key]
            except TypeError:  # if d is not accessible though key, it will raise TypeError
                raise KeyError()
        return d


def wrap_key_matcher(key_matcher, base_key=()):
    if isinstance(key_matcher, EntryMatcher):
        return key_matcher
    if type(key_matcher) not in (list, tuple):
        key_matcher = (key_matcher,)
    return KeyPathMatcher(tuple(base_key) + tuple(key_matcher))


class HasEntry(Matcher):
    def __init__(self, key_matcher, value_matcher):
        self.key_matcher = key_matcher
        self.value_matcher = value_matcher

    def short_description(self, conjugate=False):
        ret = '%s entry %s' % (to_have(conjugate), self.key_matcher.description())
        if self.value_matcher:
            ret += " that " + self.value_matcher.short_description(conjugate=True)
        return ret

    def description(self, conjugate=False):
        ret = '%s entry %s' % (to_have(conjugate), self.key_matcher.description())
        if self.value_matcher:
            ret += " that " + self.value_matcher.description(conjugate=True)
        return ret

    def matches(self, actual):
        try:
            value = self.key_matcher.get_entry(actual)
        except KeyError:
            return match_failure('No entry %s' % self.key_matcher.description())

        if self.value_matcher:
            return self.value_matcher.matches(value)
        else:
            return match_success(got_value(value))


def has_entry(key_matcher, value_matcher=None):
    """
    Test if dict has a <key> entry whose value matches (optional) value_matcher.
    Key entry can a standard dict key or a list of key where each element represent a
    level of depth of the dict (when dict are imbricated)
    """
    return HasEntry(
        wrap_key_matcher(key_matcher),
        is_(value_matcher) if value_matcher is not None else None
    )



'''
Created on Mar 12, 2017

@author: nicolas
'''

from __future__ import print_function

import os.path as osp

import lemoncheesecake
from lemoncheesecake.project import find_project_dir, find_project_file, load_project_from_file
from lemoncheesecake.reporting import get_available_backends
from lemoncheesecake.reporting.reportdir import DEFAULT_REPORT_DIR_NAME
from lemoncheesecake.filter import make_run_filter, filter_suites
from lemoncheesecake.exceptions import UserError, ProjectError

LEMONCHEESECAKE_VERSION = "lemoncheesecake version %s" % lemoncheesecake.__version__


def filter_suites_from_cli_args(suites, cli_args):
    filtr = make_run_filter(cli_args)
    if filtr.is_empty():
        return suites

    suites = filter_suites(suites, filtr)
    if len(suites) == 0:
        raise UserError("The filter does not match any test")

    return suites


def get_suites_from_project(project, cli_args):
    suites = project.get_suites()
    if all(suite.is_empty() for suite in suites):
        raise UserError("No test is defined in your lemoncheesecake project.")

    return filter_suites_from_cli_args(suites, cli_args)


def auto_detect_reporting_backends():
    project_filename = find_project_file()
    if project_filename is None:
        return get_available_backends()

    try:
        project = load_project_from_file(project_filename)
        return project.get_all_reporting_backends()
    except ProjectError:
        return get_available_backends()


def add_report_path_cli_arg(cli_parser):
    cli_parser.add_argument("report_path", nargs='?', help="Report file or directory")


def get_report_path(cli_args):
    if cli_args.report_path is None:
        project_dirname = find_project_dir()
        if project_dirname is None:
            raise UserError("Cannot find project")
        return osp.join(project_dirname, DEFAULT_REPORT_DIR_NAME)
    else:
        return cli_args.report_path



"""
DEPRECATED

lemoncheesecake.checkers API is DEPRECATED.
Please use lemoncheesecake.matching API instead.
"""

'''
Created on Jan 24, 2016

@author: nicolas
'''

from lemoncheesecake.runtime import get_runtime
from lemoncheesecake.exceptions import AbortTest
from lemoncheesecake.utils import IS_PYTHON3

import sys

CHECKER_OBJECTS = {}
BASE_CHECKER_NAMES = [ ]

def check(description, outcome, details=None):
    get_runtime().log_check(description, outcome, details)
    return outcome

class Check:
    assertion = False
    always_display_details = False
    comparator = None
    description_prefix = "Check that"
    comparator_label = None
    value_type = None
    doc_func_args = "name, actual, expected"

    def __init__(self, name, assertion=False, value_type=None):
        self.name = name
        self.assertion = assertion
        self.value_type = value_type

    def handle_check_outcome(self, outcome):
        if self.assertion:
            if outcome:
                return
            else:
                raise AbortTest("previous assertion was not fulfilled")
        else:
            return outcome

    def __call__(self, name, actual, expected):
        outcome = self.compare(name, actual, expected)
        return self.handle_check_outcome(outcome)

    def compare(self, name, actual, expected):
        description = self.format_description(name, expected)
        if self.value_type and expected != None:
            if type(actual) != self.value_type:
                return check(description, False, self.format_details(actual))
        outcome = self.comparator(actual, expected)
        details = None
        if not outcome or self.always_display_details:
            details = self.format_details(actual)
        return check(description, outcome, details)

    def format_actual_value(self, value):
        return "%s" % str(value)
    format_expected_value = format_actual_value

    def format_description(self, name, expected):
        description = u"{prefix} {name} {comparator} {expected}".format(
            prefix=self.description_prefix, name=name,
            comparator=self.comparator_label, expected=self.format_expected_value(expected)
        )
        if self.value_type and expected != None:
            description += " (%s)" % self.value_type.__name__
        return description

    def format_details(self, actual):
        details = "Got %s" % self.format_actual_value(actual)
        if self.value_type and actual != None:
            details += " (%s)" % type(actual).__name__
        return details

    def build_doc_func_args(self):
        return self.doc_func_args

    def build_doc_func_ret(self):
        return self.doc_func_ret

    def build_doc_func_description(self):
        return "{prefix} actual {comparator} expected".format(
            prefix=self.description_prefix, comparator=self.comparator_label
        )

    def build_doc(self, func_name):
        doc = "%s(%s)" % (func_name, self.build_doc_func_args())
        if not self.assertion:
            doc += " -> bool"
        doc += "\n\n%s" % self.build_doc_func_description()
        if self.value_type:
            doc += " (value's type must be %s)" % self.value_type.__name__
        doc += "\n"
        if self.assertion:
            doc += "Raise AbortTest if the check does not succeed"
        else:
            doc += "Return True if the check succeed, False otherwise"
        return doc

def do_register_checker(name, checker_inst, assertion_inst):
    def make_func(obj, func_name):
        def func(*args, **kwargs):
            return obj(*args, **kwargs)
        func.__doc__ = obj.build_doc(func_name)
        return func

    def register_checker_object(name, inst):
        CHECKER_OBJECTS[name] = inst
        setattr(sys.modules[__name__], name, make_func(inst, name))

    register_checker_object("check_%s" % name, checker_inst)
    register_checker_object("assert_%s" % name, assertion_inst)

def register_checker(name, checker_class, value_type=None, is_base_checker=True):
    if is_base_checker:
        global BASE_CHECKER_NAMES
        BASE_CHECKER_NAMES.append(name)

    checker_inst = checker_class(name, value_type=value_type)
    assertion_inst = checker_class(name, assertion=True, value_type=value_type)
    do_register_checker(name, checker_inst, assertion_inst)

def checker(name, value_type=None, is_base_checker=True):
    def wrapper(klass):
        klass.name = name
        register_checker(name, klass, value_type, is_base_checker=is_base_checker)
        return klass
    return wrapper

def get_checker_function(name):
    return getattr(sys.modules[__name__], "check_%s" % name)

def get_assertion_function(name):
    return getattr(sys.modules[__name__], "assert_%s" % name)

def get_checker_object(name):
    return CHECKER_OBJECTS["check_%s" % name]

def get_assertion_object(name):
    return CHECKER_OBJECTS["assert_%s" % name]

################################################################################
# Equality / non-equality checkers
################################################################################

@checker("eq")
class CheckEq(Check):
    comparator_label = "is equal to"
    comparator = staticmethod(lambda a, e: a == e)

@checker("not_eq")
class CheckNotEq(Check):
    comparator_label = "is not equal to"
    comparator = staticmethod(lambda a, e: a != e)
    always_display_details = True

################################################################################
# Greater than and Greater than or equal checkers
################################################################################

@checker("gt")
class CheckGt(Check):
    comparator_label = "is greater than"
    comparator = staticmethod(lambda a, e: a > e)
    always_display_details = True

@checker("gteq")
class CheckGteq(Check):
    comparator_label = "is greater or equal than"
    comparator = staticmethod(lambda a, e: a >= e)
    always_display_details = True

################################################################################
# Lower than and Lower than or equal checkers
################################################################################

@checker("lt")
class CheckLt(Check):
    comparator_label = "is lower than"
    comparator = staticmethod(lambda a, e: a < e)
    always_display_details = True

@checker("lteq")
class CheckLteq(Check):
    comparator_label = "is lower or equal than"
    comparator = staticmethod(lambda a, e: a <= e)
    always_display_details = True

################################################################################
# str checkers
################################################################################

@checker("str_eq")
class CheckStrEq(CheckEq):
    format_expected_value = format_actual_value = staticmethod(lambda s: "'%s'" % s)

@checker("str_not_eq")
class CheckStrNotEq(CheckNotEq):
    format_expected_value = format_actual_value = staticmethod(lambda s: "'%s'" % s)

@checker("str_match")
class CheckStrMatchPattern(CheckStrEq):
    comparator_label = "match pattern"
    format_expected_value = staticmethod(lambda p: "'%s'" % p.pattern)
    comparator = staticmethod(lambda a, e: bool(e.match(a)))
    always_display_details = True

@checker("str_does_not_match")
class CheckStrDoesNotMatchPattern(CheckStrMatchPattern):
    comparator_label = "does not match pattern"
    comparator = staticmethod(lambda a, e: not bool(e.match(a)))

@checker("str_contains")
class CheckStrContains(CheckStrEq):
    comparator_label = "contains"
    comparator = staticmethod(lambda a, e: e in a)
    always_display_details = True

@checker("str_does_not_contain")
class CheckStrDoesNotContain(CheckStrEq):
    comparator_label = "does not contain"
    comparator = staticmethod(lambda a, e: e not in a)
    always_display_details = True

################################################################################
# Typed checkers
################################################################################

def generate_comparator_checkers_for_type(type_):
    checker_classes = CheckEq, CheckNotEq, CheckGt, CheckGteq, CheckLt, CheckLteq
    for klass in checker_classes:
        register_checker("%s_%s" % (type_.__name__, klass.name), klass, value_type=type_)

generate_comparator_checkers_for_type(int)
generate_comparator_checkers_for_type(float)

register_checker("bool_eq", CheckEq, value_type=bool)

################################################################################
# list checkers
################################################################################

@checker("list_eq")
class CheckListEq(CheckEq):
    pass

@checker("list_len")
class CheckListLen(Check):
    comparator = staticmethod(lambda a, e: len(a) == e)
    def format_description(self, name, expected):
        return u"{prefix} {name} contains {expected} elements".format(
            prefix=self.description_prefix, name=name, expected=expected
        )
    def format_details(self, actual):
        return "Got %d elements: %s" % (len(actual), actual)

@checker("list_contains")
class CheckListContains(Check):
    comparator_label = "contains elements"
    always_display_details = True

    def compare(self, name, actual, expected):
        description = self.format_description(name, expected)

        missing = list(expected)
        for elem in expected:
            if elem in actual:
                missing.remove(elem)
        if missing:
            details = "Missing elements %s in list %s" % (missing, actual)
            return check(description, False, details)
        else:
            return check(description, True, None)

@checker("choice")
class CheckChoice(Check):
    comparator_label = "is among"
    always_display_details = True
    comparator = staticmethod(lambda a, e: a in e)

################################################################################
# dict checkers
################################################################################

def register_dict_checkers(dict_checker_name_fmt, dict_checker):
    def wrapper(value_checker):
        class dict_value_checker(dict_checker):
            doc_func_args = "key, d, expected"

            def build_doc_func_description(self):
                return "{prefix} key[d] {comparator} expected".format(
                    prefix=self.description_prefix, comparator=value_checker.comparator_label
                )

            def __call__(self, *args, **kwargs):
                kwargs["value_checker"] = value_checker
                return dict_checker.__call__(self, *args, **kwargs)
        return dict_value_checker
    global BASE_CHECKER_NAMES
    for name in BASE_CHECKER_NAMES:
        klass = wrapper(get_checker_object(name))
        do_register_checker(dict_checker_name_fmt % name, klass(name), klass(name, assertion=True))

@checker("dict_has_key", is_base_checker=False)
class CheckDictHasKey(Check):
    doc_func_args = "key, d"

    def build_doc_func_description(self):
        return "{prefix} d has entry key".format(
            prefix=self.description_prefix
        )

    def __call__(self, expected_key, actual, key_label=None, show_actual=False):
        if not key_label:
            key_label = "'%s'" % expected_key
        description = u"{prefix} entry {key_label} is present".format(prefix=self.description_prefix, key_label=key_label)
        outcome = expected_key in actual
        check(description, outcome, self.format_details(actual[expected_key]) if outcome and show_actual else None)
        self.handle_check_outcome(outcome)
        return outcome

@checker("dict_has_not_key", is_base_checker=False)
class CheckDictHasNotKey(Check):
    doc_func_args = "key, d"

    def build_doc_func_description(self):
        return "{prefix} d does not have entry key".format(
            prefix=self.description_prefix
        )

    def __call__(self, expected_key, actual, key_label=None, show_actual=False):
        if not key_label:
            key_label = "'%s'" % expected_key
        description = u"{prefix} entry {key_label} is not present".format(prefix=self.description_prefix, key_label=key_label)
        outcome = expected_key not in actual
        check(description, outcome, self.format_details(actual[expected_key]) if not outcome and show_actual else None)
        self.handle_check_outcome(outcome)
        return outcome

def register_dict_has_key_typed(type_name, types, base_class, default_show_actual=True):
    class _CheckDictHasKeyTyped(base_class):
        always_display_details = True
        doc_func_args = "key, d"

        def build_doc_func_description(self):
            return "Check that key[d] has key of type %s" % type_name

        def __call__(self, expected_key, actual, key_label=None, show_actual=default_show_actual):
            if not key_label:
                key_label = "'%s'" % expected_key
            description = u"{prefix} entry {key_label} with type {typ} is present".format(
                prefix=self.description_prefix, key_label=key_label, typ=type_name
            )
            if expected_key in actual:
                outcome = type(actual[expected_key]) in types
                ret = check(description, outcome, self.format_details(actual[expected_key]) if not outcome or show_actual else None)
            else:
                ret = check(description, False, "There is no key %s" % key_label)
            return self.handle_check_outcome(ret)

    do_register_checker(
        "dict_has_%s" % type_name,
        _CheckDictHasKeyTyped(type_name, value_type=types[0]),
        _CheckDictHasKeyTyped(type_name, value_type=types[0], assertion=True)
    )

register_dict_has_key_typed("int", [int], Check)
register_dict_has_key_typed("float", [float], Check)
if IS_PYTHON3:
    register_dict_has_key_typed("str", [str], CheckStrEq)
else:
    register_dict_has_key_typed("str", [str, unicode], CheckStrEq)
register_dict_has_key_typed("bool", [bool], Check)
register_dict_has_key_typed("list", [list], Check, default_show_actual=False)
register_dict_has_key_typed("dict", [dict], Check, default_show_actual=False)

@checker("dict_value", is_base_checker=False)
class CheckDictValue(Check):
    doc_func_args = "key, d, expected, value_checker"

    def build_doc_func_description(self):
        return "Check key[d] against expected using value_checker"

    def __call__(self, expected_key, actual, expected_value, value_checker, key_label=None):
        if not key_label:
            key_label = "'%s'" % expected_key
        if expected_key in actual:
            ret = value_checker(key_label, actual[expected_key], expected_value)
        else:
            check(value_checker.format_description(expected_key, expected_value), False,
                  "There is no key %s" % key_label)
            ret = False
        return self.handle_check_outcome(ret)

register_dict_checkers("dictval_%s", CheckDictValue)

###
# Build symbol list for wild import
###

__all__ = ("check",)
for symbol in dir():
    if symbol.startswith("check_") or symbol.startswith("assert_"):
        __all__ = __all__ + (symbol,)



from __future__ import print_function

from termcolor import colored

from lemoncheesecake.utils import get_status_color
from lemoncheesecake.cli.command import Command
from lemoncheesecake.cli.utils import auto_detect_reporting_backends
from lemoncheesecake.reporting import load_report
from lemoncheesecake.filter import add_report_filter_cli_args, make_report_filter, \
    filter_suites
from lemoncheesecake.testtree import flatten_tests, find_test
from lemoncheesecake.exceptions import CannotFindTreeNode, UserError

ORDERED_STATUSES = "failed", "skipped", "disabled", "passed"


class Diff:
    def __init__(self):
        self.added = []
        self.removed = []
        self.status_changed = {}

    def is_empty(self):
        return len(self.added) + len(self.removed) + len(self.status_changed) == 0


def compute_diff(old_suites, new_suites):
    diff = Diff()

    new_tests = {test.path: test for test in flatten_tests(new_suites)}
    for old_test in flatten_tests(old_suites):
        # handle removed tests
        try:
            new_test = find_test(new_suites, old_test.path)
        except CannotFindTreeNode:
            diff.removed.append(old_test)
            continue

        # handle status-changed tests
        if new_test.status != old_test.status:
            if old_test.status not in diff.status_changed:
                diff.status_changed[old_test.status] = {}
            if new_test.status not in diff.status_changed[old_test.status]:
                diff.status_changed[old_test.status][new_test.status] = []
            diff.status_changed[old_test.status][new_test.status].append(new_test)

        del new_tests[new_test.path]

    # handle added tests
    diff.added.extend(new_tests.values())

    return diff


def display_diff_type(title, test_entries):
    if len(test_entries) == 0:
        return

    print("%s (%d):" % (colored(title, attrs=["bold"]), len(test_entries)))
    for test_entry in test_entries:
        print("- %s" % test_entry)
    print()


def render_test_with_status(test):
    return "%s (%s)" % (test.path, colored(test.status, get_status_color(test.status)))


def render_test_with_status_changed(test, old_status):
    return "%s (%s => %s)" % (
        test.path,
        colored(old_status, get_status_color(old_status)),
        colored(test.status, get_status_color(test.status))
    )


def flatten_test_status_changed(status_changed):
    for old_status in ORDERED_STATUSES:
        for new_status in ORDERED_STATUSES:
            for test in status_changed.get(old_status, {}).get(new_status, []):
                yield test, old_status


def display_diff(diff):
    if diff.is_empty():
        print("There is no difference between the two reports.")
    else:
        display_diff_type(
            "Added tests", [render_test_with_status(test) for test in diff.added]
        )
        display_diff_type(
            "Removed tests", [render_test_with_status(test) for test in diff.removed]
        )
        display_diff_type(
            "Status changed",
            [
                render_test_with_status_changed(test, old_status)
                    for test, old_status in flatten_test_status_changed(diff.status_changed)
            ]
        )


class DiffCommand(Command):
    def get_name(self):
        return "diff"

    def get_description(self):
        return "Display differences between two reports"

    def add_cli_args(self, cli_parser):
        cli_parser.add_argument("old_report_path", help="Old report path")
        cli_parser.add_argument("new_report_path", help="New report path")
        add_report_filter_cli_args(cli_parser, no_positional_argument=True)

    def run_cmd(self, cli_args):
        reporting_backends = auto_detect_reporting_backends()

        old_report = load_report(cli_args.old_report_path, reporting_backends)
        new_report = load_report(cli_args.new_report_path, reporting_backends)
        filtr = make_report_filter(cli_args)

        old_suites = filter_suites(old_report.suites, filtr)
        new_suites = filter_suites(new_report.suites, filtr)

        if len(old_suites) == 0 and len(new_suites) == 0:
            raise UserError("The filter does not match any test on both reports")

        diff = compute_diff(old_suites, new_suites)
        display_diff(diff)

        return 0
